#include "smat.h"
#include <cstdlib>
#include <algorithm>
#include <iostream>
#include <iterator>
#include <map>
#include <sstream>
#include <utility>
#include <vector>
#ifdef linux
#include <getopt.h>
#endif
int accumulate_stride_1(std::vector<int>& v, int stride)
{
  int sum = 0;
  for (size_t i = 0 ; i < v.size(); i += stride)
    sum += v[i];
  return sum;
}
int accumulate_stride_2(std::vector<int>::iterator first,
                        std::vector<int>::iterator last, int stride)
{
  int sum = 0;
  for ( ; first < last; first += stride)
    sum += *first;
  return sum;
}
template <typename T, typename U>
int accumulate_stride_3(T first, T last, U stride)
{
  int sum = 0;
  for ( ; first < last; first += stride)
    sum += *first;
  return sum;
}
template <typename T, typename U>
typename std::iterator_traits<T>::value_type
accumulate_stride_4(T first, T last, U stride)
{
  typename std::iterator_traits<T>::value_type sum(*first);
  first += stride;
  for ( ; first < last; first += stride)
    sum = sum + *first;
  return sum;
}
template <typename T>
typename std::iterator_traits<T>::value_type
accumulate_stride_final(T first, T last,
  typename std::iterator_traits<T>::difference_type stride)
{
  typename std::iterator_traits<T>::value_type sum(*first);
  first += stride;
  for ( ; first < last; first += stride)
    sum = sum + *first;
  return sum;
}

int main()
{
  int x[5] = { 3, 6, 7, 2, 4 };
  float y[5] = { 3, 6, 7, 2, 3.5 };
  std::vector<int> v(&x[0], &x[5]);
  typedef std::vector<int>::iterator iter;

  std::cerr << accumulate_stride_1(v, 2) << '\n';
  std::cerr << accumulate_stride_2(v.begin(), v.end(), 2) << '\n';
  std::cerr << accumulate_stride_3(&x[0], &x[5], 2) << '\n';
  std::cerr << accumulate_stride_4(&y[0], &y[5], 2) << '\n';
  std::cerr << accumulate_stride_4(smat<iter>(v.begin()),
                                   smat<iter>(v.end()), 2) << '\n';
  std::cerr << accumulate_stride_final(smat<iter>(v.begin()),
                                       smat<iter>(v.end()), 2) << '\n';
}
